@@ -5,10 +5,12 @@ package Interchange6::Cart;
 use strict;
 use Carp;
 use DateTime;
+use Interchange6::Cart::Cost;
 use Interchange6::Cart::Product;
 use Scalar::Util 'blessed';
 use Try::Tiny;
 use Moo;
+use MooseX::CoverableModifiers;
 use MooX::HandlesVia;
 use Interchange6::Types;
 use Interchange6::Hook;
@@ -24,14 +26,16 @@ use constant CART_DEFAULT => 'main';
 
 has costs => (
     is          => 'rwp',
-    isa         => ArrayRef [HashRef],
+    isa         => ArrayRef [ InstanceOf ['Interchange::Cart::Cost'] ],
     default     => sub { [] },
     handles_via => 'Array',
     handles     => {
-        clear_cost => 'clear',
-        cost_get   => 'get',
-        _cost_push => 'push',
-        get_costs  => 'elements',
+        clear_cost  => 'clear',
+        clear_costs => 'clear',
+        cost_get    => 'get',
+        cost_count  => 'count',
+        _cost_push  => 'push',
+        get_costs   => 'elements',
     },
     init_arg => undef,
 );
@@ -85,7 +89,6 @@ has products => (
         _product_set   => 'set',
     },
     reader   => 'get_products',
-    writer   => 'set_products',
     init_arg => undef,
 );
 
@@ -195,14 +198,15 @@ around clear => sub {
 
 around set_name => sub {
     my ( $orig, $self ) = ( shift, shift );
+    my $new_name = $_[0];
     my $ret;
 
     $self->clear_error;
 
     my $old_name = $self->get_name;
 
-    # run hook before clearing the cart
-    $self->execute_hook( 'before_cart_rename', $self, $old_name, $_[0] );
+    # run hook before renaming the cart
+    $self->execute_hook( 'before_cart_rename', $self, $old_name, $new_name );
     return if $self->has_error;
 
     # fire off the rename
@@ -211,7 +215,7 @@ around set_name => sub {
     $self->_set_last_modified( DateTime->now );
 
     # run hook after clearing the cart
-    $self->execute_hook( 'after_cart_rename', $self, $old_name, $_[0] );
+    $self->execute_hook( 'after_cart_rename', $self, $old_name, $new_name );
 
     return $self->get_name;
 };
@@ -225,8 +229,11 @@ sub add {
 
     $self->clear_error;
 
-    unless ( blessed($product) && $product->isa('Interchange6::Cart::Product') )
-    {
+    if ( blessed($product) ) {
+        die "product argument is not an Interchange6::Cart::Product"
+          unless ( $product->isa('Interchange6::Cart::Product') );
+    }
+    else {
 
         # we got a hash(ref) rather than an Product
 
@@ -297,12 +304,33 @@ sub add {
 }
 
 sub apply_cost {
-    my ( $self, %args ) = @_;
+    my $self = shift;
+    my $cost = $_[0];
 
-    $self->_cost_push( \%args );
+    die "argument to apply_cost undefined" unless defined($cost);
+
+    if ( blessed($cost) ) {
+        die("Supplied cost not an Interchange6::Cart::Cost : " . ref($cost))
+          unless $cost->isa('Interchange6::Cart::Cost');
+    }
+    else {
+        if ( @_ % 2 ) {
+
+            # a hashref or obj
+            $cost = @_;
+        }
+        else {
+
+            # hash
+            $cost = {@_};
+        }
+        $cost = Interchange6::Cart::Cost->new( $cost );
+    }
+
+    $self->_cost_push( $cost );
 
     # clear cache for total if this is not an inclusive cost
-    $self->clear_total unless $args{inclusive};
+    $self->clear_total unless $cost->inclusive;
 }
 
 sub cost {
@@ -319,16 +347,22 @@ sub cost {
 
             # cost by name
             for my $c ( $self->get_costs ) {
-                if ( $c->{name} eq $loc ) {
+                if ( $c->name eq $loc ) {
                     $cost = $c;
                 }
             }
         }
     }
+    else {
+        die "Either position or name required as argument to cost";
+    }
 
     if ( defined $cost ) {
         $ret = $self->_calculate( $self->subtotal, $cost, 1 );
     }
+    else {
+        die "Bad argument to cost: " . $loc;
+    }
 
     return $ret;
 }
@@ -531,15 +565,15 @@ sub _calculate {
     $sum = 0;
 
     for my $calc (@$cost_ref) {
-        if ( $calc->{inclusive} && !$display ) {
+        if ( $calc->inclusive && !$display ) {
             next;
         }
 
-        if ( $calc->{relative} ) {
-            $sum += $subtotal * $calc->{amount};
+        if ( $calc->relative ) {
+            $sum += $subtotal * $calc->amount;
         }
         else {
-            $sum += $calc->{amount};
+            $sum += $calc->amount;
         }
     }
 
@@ -651,6 +685,10 @@ Removes all products from the cart.
 
 Removes all the costs previously applied (using apply_cost). Used typically if you have free shipping or something similar, you can clear the costs.
 
+=head2 clear_costs
+
+Alias for clear_cost
+
 =head2 cost
 
 Returns particular cost by position or by name.
@@ -671,6 +709,10 @@ Returns the cost that was first applied to subtotal. By increasing the number yo
 
 Returns an array of all costs associated with the cart. Costs are ordered according to the order they were applied.
 
+=head2 cost_count
+
+Returns the number of different costs applied to the shopping cart.
+
 =head2 count
 
 Returns the number of different products in the shopping cart. If you have 5 apples and 6 pears it will return 2 (2 different products).
